/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.registry.jdbc.dao;

import org.wso2.registry.Resource;
import org.wso2.registry.RegistryConstants;
import org.wso2.registry.jdbc.DatabaseConstants;
import org.wso2.registry.versions.VersionHandler;

import java.sql.*;
import java.util.*;

public class ResourceDAO {

    public Resource get(String path, Connection conn) throws SQLException {

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.PATH=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setString(1, path);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            resource.setLastUpdaterUserName(result.getString(DatabaseConstants.LAST_UPDATER_FIELD));
            resource.setLastModified(result.getTimestamp(DatabaseConstants.LAST_UPDATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
            resource.setContent(result.getBytes(DatabaseConstants.CONTENT_FIELD));
        }

        // get resource chidren

        if (resource != null && resource.isDirectory()) {

            String childrenSQL = "SELECT A.PATH FROM ARTIFACTS A, CHILDREN C WHERE A.AID=C.CHILD_ID AND C.PARENT_ID=?";
            PreparedStatement s2 = conn.prepareStatement(childrenSQL);
            s2.setLong(1, resource.getId());
            ResultSet childrenResults = s2.executeQuery();

            List childrenList = new ArrayList();
            while (childrenResults.next()) {
                childrenList.add(childrenResults.getString(DatabaseConstants.PATH_FIELD));
            }

            String[] children =
                    (String[]) childrenList.toArray(new String[childrenList.size()]);
            resource.setContent(children);
        }

        // get resource properties

        if (resource != null) {

            String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
            PreparedStatement s3 = conn.prepareStatement(propSQL);
            s3.setLong(1, resource.getId());
            ResultSet propResults = s3.executeQuery();

            Properties props = new Properties();
            while (propResults.next()) {
                String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
                String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
                props.put(key, value);
            }
            resource.setProperties(props);
        }

        return resource;
    }

    public Resource get(String path, long versionNumber, Connection conn) throws SQLException {

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.PATH=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setString(1, path);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(path + "?v=" + versionNumber);
            resource.setParentPath(VersionHandler.getParentPath(resource.getId(), versionNumber, conn));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            //resource.setLastUpdaterUserName(result.getString(DatabaseConstants.LAST_UPDATER_FIELD));
            //resource.setLastModified(result.getTimestamp(DatabaseConstants.LAST_UPDATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
            //resource.setContent(result.getBytes(DatabaseConstants.CONTENT_FIELD));
        }

        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, resource.getId());
        versionQuery.setLong(2, versionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            resource.setLastUpdaterUserName(versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            resource.setLastModified(versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            resource.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get resource chidren

        if (resource != null && resource.isDirectory()) {

            //String childrenSQL = "SELECT A.PATH FROM ARTIFACTS A, CHILDREN C WHERE A.AID=C.CHILD_ID AND C.PARENT_ID=?";
            //PreparedStatement s2 = conn.prepareStatement(childrenSQL);
            //s2.setLong(1, resource.getId());
            //ResultSet childrenResults = s2.executeQuery();
            //
            //List childrenList = new ArrayList();
            //while (childrenResults.next()) {
            //    childrenList.add(childrenResults.getString(DatabaseConstants.PATH_FIELD));
            //}
            //
            //String[] children =
            //        (String[]) childrenList.toArray(new String[childrenList.size()]);

            String[] children = VersionHandler.
                    getChildrenPath(resource.getId(), versionNumber, conn);
            resource.setContent(children);
        }

        // get resource properties

        if (resource != null) {

            String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
            PreparedStatement s3 = conn.prepareStatement(propSQL);
            s3.setLong(1, resource.getId());
            ResultSet propResults = s3.executeQuery();

            Properties props = new Properties();
            while (propResults.next()) {
                String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
                String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
                props.put(key, value);
            }
            resource.setProperties(props);
        }

        return resource;
    }

    public Resource getArtifactByID(long artifactID, long versionNumber, Connection conn) throws SQLException {

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.AID=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setLong(1, artifactID);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (versionNumber == -1) {
            versionNumber = VersionHandler.getLatestVerison(artifactID, conn);
        }        
        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, resource.getId());
        versionQuery.setLong(2, versionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            resource.setLastUpdaterUserName(versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            resource.setLastModified(versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            resource.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get resource chidren

        if (resource.isDirectory()) {

            String[] children = VersionHandler.
                    getChildrenPath(resource.getId(), versionNumber, conn);
            resource.setContent(children);
        }

        // get resource properties

            String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
            PreparedStatement s3 = conn.prepareStatement(propSQL);
            s3.setLong(1, resource.getId());
            ResultSet propResults = s3.executeQuery();

            Properties props = new Properties();
            while (propResults.next()) {
                String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
                String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
                props.put(key, value);
            }
            resource.setProperties(props);
        return resource;
    }

    public Resource getLatestVersion(String path, Connection conn) throws SQLException {

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.PATH=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setString(1, path);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            resource.setParentPath(getParentPath(resource.getPath()));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            //resource.setLastUpdaterUserName(result.getString(DatabaseConstants.LAST_UPDATER_FIELD));
            //resource.setLastModified(result.getTimestamp(DatabaseConstants.LAST_UPDATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
            //resource.setContent(result.getBytes(DatabaseConstants.CONTENT_FIELD));
        }

        long latestVersionNumber = VersionHandler.getLatestVerison(resource.getId(), conn);

        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, resource.getId());
        versionQuery.setLong(2, latestVersionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            resource.setLastUpdaterUserName(versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            resource.setLastModified(versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            resource.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get resource chidren

        if (resource.isDirectory()) {

            //String childrenSQL = "SELECT A.PATH FROM ARTIFACTS A, CHILDREN C WHERE A.AID=C.CHILD_ID AND C.PARENT_ID=?";
            //PreparedStatement s2 = conn.prepareStatement(childrenSQL);
            //s2.setLong(1, resource.getId());
            //ResultSet childrenResults = s2.executeQuery();
            //
            //List childrenList = new ArrayList();
            //while (childrenResults.next()) {
            //    childrenList.add(childrenResults.getString(DatabaseConstants.PATH_FIELD));
            //}
            //
            //String[] children =
            //        (String[]) childrenList.toArray(new String[childrenList.size()]);

            String[] children = VersionHandler.
                    getLatestChildrenPath(resource.getId(), latestVersionNumber, conn);
            resource.setContent(children);
        }

        // get resource properties
        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, resource.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        resource.setProperties(props);
        return resource;
    }

    /**
     * Adds the given resource to the database.
     *
     * @param path Path of the new resource
     * @param resource New resource
     * @param conn : Connection to DB
     */
    public void add (String path, Resource resource, Connection conn) throws SQLException {

        // first add the resource data into the artifacts table

        String sql = "INSERT INTO ARTIFACTS (PATH, MEDIA_TYPE, DIRECTORY, STATE, CREATED_TIME, AUTHOR, LAST_UPDATER, LAST_UPDATED_TIME,  DESCRIPTION, CONTENT)" +
                "VALUES (?,?,?,?,?,?,?,?,?,?)";

        long now = System.currentTimeMillis();

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, resource.getPath());
        s.setString(2, resource.getMediaType());
        s.setInt(3, resource.isDirectory()? 1: 0);
        s.setInt(4, RegistryConstants.ACTIVE_STATE);
        s.setTimestamp(5, new Timestamp(now));
        s.setString(6, resource.getAuthorUserName());
        s.setString(7, resource.getLastUpdaterUserName());
        s.setTimestamp(8, new Timestamp(now));
        s.setString(9, resource.getDescription());

        if (resource.getContent() != null && resource.getContent() instanceof byte[]) {
            s.setBytes(10, (byte[]) resource.getContent());
        } else {
            s.setBytes(10, null);
        }

        s.executeUpdate();        

        long newArtifactID = this.getArtifactID(resource.getPath(), conn);

        // now add resource properties to the properties table
        addProperties(newArtifactID, resource.getProperties(), conn);
    }

    public void update(String path, Resource resource, Connection conn) throws SQLException{

        // first update the artifacts table

        String sql = "UPDATE ARTIFACTS SET MEDIA_TYPE=?, DIRECTORY=?, STATE=?, AUTHOR=?, LAST_UPDATER=?, LAST_UPDATED_TIME=?, DESCRIPTION=?, CONTENT=? WHERE PATH=?";

        long now = System.currentTimeMillis();

        PreparedStatement s = conn.prepareStatement(sql);

        s.setString(1, resource.getMediaType());
        s.setInt(2, resource.isDirectory()? 1: 0);
        s.setInt(3, resource.getState());
        s.setString(4, resource.getAuthorUserName());
        s.setString(5, resource.getLastUpdaterUserName());
        s.setTimestamp(6, new Timestamp(now));
        s.setString(7, resource.getDescription());

        if (resource.getContent() != null && resource.getContent() instanceof byte[]) {
            s.setBytes(8, (byte[]) resource.getContent());
        } else {
            s.setBytes(8, null);
        }

        s.setString(9, path);

        s.executeUpdate();

        // remove existing properties for the updated artfact
        removeProperties(resource.getId(), conn);

        // now add the updated properties to the properties table
        addProperties(resource.getId(), resource.getProperties(), conn);
    }

    public void delete(String path, Connection conn) throws SQLException {

        String sql = "DELETE FROM ARTIFACTS WHERE PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);
        s.executeUpdate();
    }

    public void deleteByID(long id, Connection conn) throws SQLException {

        String sql = "DELETE FROM ARTIFACTS WHERE AID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, id);
        s.executeUpdate();
    }

    public void markDeleted(String path, Connection conn) throws SQLException {

        String sql = "UPDATE ARTIFACTS SET STATE=? WHERE PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setInt(1, RegistryConstants.DELETED_STATE);
        s.setString(2, path);
        s.executeUpdate();
    }

    public void addChild(long parentID, long childID, Connection conn) throws SQLException{

        String sql = "INSERT INTO CHILDREN (PARENT_ID, CHILD_ID) VALUES (?,?)";

        PreparedStatement statement = conn.prepareStatement(sql);

        statement.setLong(1, parentID);
        statement.setLong(2, childID);

        statement.executeUpdate();
    }

    public void removeChildRelationship(long childID, Connection conn) throws SQLException {

        String sql = "DELETE FROM CHILDREN WHERE CHILD_ID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, childID);
        s.executeUpdate();
    }

    public List getChildren(long parentID, Connection conn) throws SQLException {

        List childList = new ArrayList();

        String sql = "SELECT A.AID, A.PATH, A.DIRECTORY FROM ARTIFACTS A, CHILDREN C WHERE A.AID=C.CHILD_ID AND C.PARENT_ID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, parentID);

        ResultSet result = s.executeQuery();
        while (result.next()) {
            Resource child = new Resource();
            child.setId(result.getLong(DatabaseConstants.AID_FIELD));
            child.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            child.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            childList.add(child);
        }

        return childList;
    }

    public List getChildren(long artifcatID, long versionNumber,
                                           Connection connection) throws SQLException {

        if (versionNumber == -1) {
            versionNumber = VersionHandler.getLatestVerison(artifcatID, connection);
        }
        
        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList childList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            String SQL = "SELECT AID, PATH, DIRECTORY FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);

            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                Resource child = new Resource();
                child.setId(pathResult.getLong(DatabaseConstants.AID_FIELD));
                child.setPath(pathResult.getString(DatabaseConstants.PATH_FIELD));
                child.setDirectory(pathResult.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
                childList.add(child);
            }
        }
        return childList;
    }

    public void addProperties(long artifactID, Properties props, Connection conn)
            throws SQLException {

        String propSQL = "INSERT INTO PROPERTIES (AID, PKEY, PVALUE) VALUES (?,?,?)";
        PreparedStatement s2 = conn.prepareStatement(propSQL);

        if (props != null) {

            Iterator i = props.keySet().iterator();
            while (i.hasNext()) {
                String key = (String) i.next();
                String value = (String) props.get(key);

                s2.setLong(1, artifactID);
                s2.setString(2, key);
                s2.setString(3, value);

                s2.executeUpdate();
                s2.clearParameters();
            }
        }
    }

    public void removeProperties(long artifactID, Connection conn) throws SQLException {

        String removePropsSQL = "DELETE FROM PROPERTIES WHERE AID=?";
        PreparedStatement s2 = conn.prepareStatement(removePropsSQL);
        s2.setLong(1, artifactID);
        s2.executeUpdate();
    }

    public boolean artifactExist(String path, Connection conn) throws SQLException {

        String sql = "SELECT COUNT(A.AID) FROM ARTIFACTS A WHERE A.PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);

        ResultSet result = s.executeQuery();
        long count = 0;
        if (result.next()) {
            count = result.getLong(1);
        }

        return count == 1;
    }

    public long getArtifactID(String path, Connection conn) throws SQLException {

        String sql = "SELECT A.AID FROM ARTIFACTS A WHERE A.PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);

        ResultSet result = s.executeQuery();
        if (result.next()) {
            return result.getLong(DatabaseConstants.AID_FIELD);
        }

        return -1;
    }

    private String getParentPath(String path) {
        String[] parts = path.split(RegistryConstants.PATH_SEPARATOR);
        String parentPath = "";
        if (parts.length > 2) {
            for (int i = 1; i < parts.length - 1; i++) {
                parentPath = parentPath + "/" + parts[i];

            }
        } else {
            parentPath = "/";
        }
        return parentPath;
    }
}
